/* Version 1.0 (Beta Test)
 * Copyright (c) Rui Zhang, 2007
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The code will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - All bug fixes will be returned to rui@csse.unimelb.edu.au for
 *    inclusion in future releases
 */

/*************************************************************************
*
* File : bt_insert.c
*
* Modules:
* BT_InsertR()
*
*
* Version 1
* Coded by: Dai Haoyu		Modified by:
* Date: 12/9/97			Date modified: 
*************************************************************************/

#include "btree.h"

extern int got;

/* function prototype */
/* internal */
int BT_InsertR(BTREE *, const long, const KEYTYPE, RECORD_TYPE *, INODE *, 
               INODE **, KEYTYPE *);

/* external */
extern int BT_IOReadLNode(BTREE *, PTR, LNODE *);
extern int BT_IOReadINode(BTREE *, PTR, INODE *);
extern int BT_IOWriteLNode(BTREE *, PTR, LNODE *);
extern int BT_IOWriteINode(BTREE *, PTR, INODE *);
extern PTR BT_IOWriteNewLNode(BTREE *, LNODE *);
extern PTR BT_IOWriteNewINode(BTREE *, INODE *);
extern int BT_NewLNode(LNODE **);
extern int BT_NewINode(INODE **);
extern int BT_NewLEntry(LENTRY **);
extern int BT_InsertLNode(BTREE *, LENTRY *, LNODE *, LNODE **);
extern int BT_InsertINode(PTR, KEYTYPE, PTR, INODE *, INODE **, KEYTYPE *);
extern PTR BT_SelectChild(INODE *, KEYTYPE);
extern int BT_VerifyLNode(LNODE *);
extern int BT_VerifyINode(INODE *);
extern void BT_PrintLNode(LNODE *);
extern void BT_PrintINode(INODE *);

extern int BT_InsertNewINode(INODE *, PTR, KEYTYPE, PTR);


/***************************************************************************

* BT_InsertR():
*   A recursive routine to insert a reocrd into the B+ tree
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if no split occurs at this level else FALSE and newNode contains
*   the newly generated node after splitting and middleKey contains the
*   key value which is to be moved up one level
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   If split occurs at this level, the new node will be passed out through
*   the last argument
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_InsertR(BTREE *btree,
               const long recid, 
               const KEYTYPE key, 
               RECORD_TYPE *rd,
               INODE *oriNode, 
               INODE **newNode,
               KEYTYPE *middleKey
) {
  PTR child, newChild;
  KEYTYPE newKey, 
          splitKey;
  LNODE *lnode ,
        *newLnode;
  INODE *inode,
        *newInode;
  LENTRY *newEntry;
  int k;

  if (oriNode->Level == 0) { /* the leaf level of INODE
                              * directly above LNODE level */
    lnode = NULL;
    newLnode = NULL;
    newEntry = NULL;
    BT_NewLNode(&lnode);
    BT_NewLEntry(&newEntry);
    newEntry->Key = key;
    newEntry->Offset = recid;
    for (k=0; k<D; k++) {
       newEntry->ver[k]=rd->fields[k];
    }

    child = BT_SelectChild(oriNode, key);
    if (child == -1) { /* the oriNode is empty, in that case it must be the
                        * root node and the tree is empty */
      if (BT_InsertLNode(btree, newEntry, lnode, &newLnode)){
      /* no splitting */
        BT_NewLNode(&newLnode);
        lnode->Next = -1L;
        newLnode->Next = BT_IOWriteNewLNode(btree, lnode);
        btree->Leaf_Head = BT_IOWriteNewLNode(btree, newLnode);
        BT_InsertNewINode(oriNode, btree->Leaf_Head, key, newLnode->Next);
        
        if (!BT_VerifyINode(oriNode)) {
          BT_PrintINode(oriNode);
        }
        
        free(lnode);
        free(newLnode);
        free(newEntry);
        return(TRUE);
      } else {  /* splitting occurs */
        printf("BT_InsertR : somthing wrong ");
      }
    } else { /* normal insertion */
      BT_IOReadLNode(btree, child, lnode);
      if (BT_InsertLNode(btree, newEntry, lnode, &newLnode)){
      /* no split at LNODE level */
        BT_IOWriteLNode(btree, child, lnode);
        free(lnode);
        free(newEntry);
        return (TRUE);
      } else { /* splitting occurs at LNODE level */
        newLnode->Next = lnode->Next;
        newChild = BT_IOWriteNewLNode(btree, newLnode);
        lnode->Next = newChild;
        BT_IOWriteLNode(btree, child, lnode);
        newKey = newLnode->Entries[0].Key;
        free(lnode);
        free(newLnode);
        free(newEntry);
        return (BT_InsertINode(newChild, newKey, child, oriNode, 
                               newNode, middleKey));
      }
    }
  } else {  /* the non-leaf level of INODE */
    child = BT_SelectChild(oriNode, key);
    BT_NewINode(&inode);
    BT_IOReadINode(btree, child, inode);
    if (BT_InsertR(btree, recid, key, rd, inode, &newInode, &splitKey)){
    /* no split */
      if (!BT_VerifyINode(inode)) {
        BT_PrintINode(inode);
      }
      BT_IOWriteINode(btree, child, inode);
      free(inode);
      return (TRUE);
    } else {  /* split occurs */
      if (!BT_VerifyINode(inode)) {
        BT_PrintINode(inode);
      }
      if (!BT_VerifyINode(newInode)) {
        BT_PrintINode(newInode);
      }
      BT_IOWriteINode(btree, child, inode);
      newChild = BT_IOWriteNewINode(btree, newInode);
      free(inode);
      free(newInode);
      return (BT_InsertINode(newChild, splitKey, child, oriNode,
                             newNode, middleKey));
    }
  }
}
