/* Version 1.0 (Beta Test)
 * Copyright (c) Rui Zhang, 2007
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The code will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - All bug fixes will be returned to rui@csse.unimelb.edu.au for
 *    inclusion in future releases
 */

/*************************************************************************
*
* File : bt_search.c
*
* Modules:
* BT_SearchLeaf()
* BT_SearchLeaf_v2()
* BT_SearchLeafRange()
*
*
* Version 1
* Coded by: Dai Haoyu		Modified by:
* Date: 12/9/97			Date modified: 
*************************************************************************/
#include "nulllist.h"
#include "btree.h"

extern int got;
//extern int candidate;

/* function prototype */
/* internal */


int BT_SearchLeaf(BTREE *, PTR, RECORD_TYPE *,const KEYTYPE,OFFSETLIST **);
/*int BT_SearchLeaf_v2(BTREE *, PTR, const KEYTYPE, OFFSETLIST **);*/
int BT_SearchLeafRange(BTREE *, PTR, const KEYTYPE, const KEYTYPE, range *,
                       OFFSETLIST **);
int BT_InsertSrhResult(BTREE *, OFFSETLIST *);

/* external */
extern int BT_IOReadLNode(BTREE *, PTR, LNODE *);
extern int BT_NewLNode(LNODE **);
extern int NL_NodeFull(const NL_NODE *);
extern void NL_AddToNode(NL_NODE *, const long);
extern int NL_NewNode(NL_NODE **);
extern void NL_InitNode(NL_NODE *);
extern int NL_IOReadNode(FILE *, const long, NL_NODE *);
extern int NL_IOWriteNode(FILE *, const long, NL_NODE *);
extern int NL_IOWriteNewNode(FILE *, NL_NODE *);

extern int BT_NewONode(ONODE **);
extern int BT_IOReadONode(BTREE *, PTR, ONODE *);
extern void InsertList(OFFSETLIST **, const long);

/***************************************************************************
* BT_SearchLeaf():
*   Routine to search a key in  LNODEs. It will try to locate the record with
*   the key value and scan through all its over-flow pages and write all the 
*   record ids to disk in format of NL_NODE. The file is btree.SrhResult.
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if something is found else FALSE
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_SearchLeaf(BTREE *btree,   /* the B+ tree to be searched from */
                  PTR currNodePtr, /* the LNODE to start searching */
                  RECORD_TYPE *rd,
                  const KEYTYPE key,  /* the key to br searched */ 
                  OFFSETLIST **resultList 
) {
  int result = FALSE;
  int i,j, k;
  LNODE *currNode = NULL;
  ONODE *oNode = NULL;
  NL_NODE *nlNode = NULL;
  PTR overPtr;
  int listLen = 0;
  int Continue = 0;
  
  BT_NewONode(&oNode);
  BT_NewLNode(&currNode);
  
  if (BT_IOReadLNode(btree, currNodePtr, currNode)) {
     for (i=0; i < currNode->NumEntry; i++) {
       if (currNode->Entries[i].Key == key) {
          result=TRUE;
          for (k=0;k<D;k++) {
             if ((currNode->Entries[i]).ver[k] != rd->fields[k]) {
                result=FALSE;
                break;
             }
          }
          if (result) {
              InsertList(resultList, currNode->Entries[i].Offset);
              listLen++;
              got++;
          }                                      
          Continue = 1;
       } else if ( currNode->Entries[i].Key > key) {
          Continue = 0;
          break;
       } else {
       }     
     } /* for */
      
     if ( Continue && currNode->overflowNum > 0) {          
          overPtr = currNode->Overflow;
          while (overPtr >= 0) {              
              BT_IOReadONode(btree, overPtr, oNode);
              for (j=0; j<oNode->NumEntry; j++) {
                 result=TRUE;
                 for (k=0;k<D;k++) {
                    if ((currNode->Entries[i]).ver[k] != rd->fields[k]) {               
                       result=FALSE;
                       break;               
                    }
                 }  
                 if (result) {           
                     InsertList(resultList, currNode->Entries[i].Offset);
                     listLen++;
                     got++;
                 }                                                  
                 overPtr = oNode->Next;
              }
          }
     }
     free(currNode);
     free(oNode);
     free(nlNode);

     if (listLen > 0) {
        return (TRUE);
     } else {
        return (FALSE);
    } 
  }   
}


/***************************************************************************
* BT_InsertSrhResult():
*   Routine to write a list of tids to disk
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if successful else FALSE
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   The list of tuple ids are written to disk in the format of NL_NODE
*
* WARNINGS:
*   The file btree.SrhResult must be opened for overwrite before calling
*   this routine
*   This routine might be called several times while writing to the same file.
*   In that case, the last node written in one invokation might not be full,
*   and the next invokation will not attempt to fill that node but to start
*   with a new node. Simplicity is the main reason for doing so.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_InsertSrhResult(BTREE *btree,   /* the btree where the result is from */
                       OFFSETLIST *list  /* the list of tids from search */
) {
  int result = TRUE;
  OFFSETLIST *current = NULL;
  NL_NODE *node;

  NL_NewNode(&node);
  NL_InitNode(node);

  current = list;

  while (current) {
    if (!NL_NodeFull(node)) {
      NL_AddToNode(node, current->Offset);
    } else {
      NL_IOWriteNewNode(btree->SrhResult, node);
      NL_InitNode(node);
      NL_AddToNode(node, current->Offset); 
    }
    current = current->next;
  }
  NL_IOWriteNewNode(btree->SrhResult, node);
  free(node);

  return result;
}

/***************************************************************************

* BT_SearchLeafRange():
*   Routine to search a range in  LNODEs. It will try to locate the record with
*   the key value and scan through all its over-flow pages and write all the 
*   record ids to disk in format of NL_NODE. The file is btree.SrhResult.
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if something is found else FALSE
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_SearchLeafRange(BTREE *btree,   /* the B+ tree to be searched from */
                       PTR currNodePtr, /* the LNODE to start searching */
                       const KEYTYPE key1,  /* lower bound */
                       const KEYTYPE key2,  /* upper bound */
                       range *rg,
                       OFFSETLIST **resultList 
) {
  int result = FALSE;
  int i,j, k;
  LNODE *currNode = NULL;
  ONODE *oNode = NULL;
  PTR   overPtr, nextPtr;
  int   listLen = 0;
/*  long preOffset;*/
  int  Continue = TRUE;
  
  BT_NewONode(&oNode);
  BT_NewLNode(&currNode);

  nextPtr = currNodePtr;

while (Continue) {
  if (Continue && BT_IOReadLNode(btree, nextPtr, currNode)) {
    for (i=0; i < currNode->NumEntry; i++) {
      if (Continue && currNode->Entries[i].Key >= key1 &&
          currNode->Entries[i].Key <= key2 ) { 
/*        printf("key: %.18f\n",currNode->Entries[i].Key); */
        //candidate++;
        result=TRUE;
        for (k=0;k<D;k++) {
           if ((currNode->Entries[i]).ver[k]<rg->start[k] 
                 || (currNode->Entries[i]).ver[k]>rg->end[k] ) {               
               result=FALSE;
               break;               
           }
        }  
   /*     preOffset = currNode->Entries[i].Offset;      */
        if (result) {           
            InsertList(resultList, currNode->Entries[i].Offset);
            listLen++;
            got++;
        }             
      } else {
         if (currNode->Entries[i].Key > key2) {
            Continue=FALSE;
         }
      } 
    } /* for */
    if (Continue && currNode->Overflow >= 0) {
        overPtr = currNode->Overflow;
        /* search in the overflow node */
        while (overPtr >= 0) {
            BT_IOReadONode(btree, overPtr, oNode);            
            for (j=0; j<oNode->NumEntry; j++) {
               /*
               if (oNode->Entries[j].Offset < preOffset) {
                   printf("BANG!!! ONODE entries not in order!!!\n");
                }
               */ 
                //candidate++;
                result=TRUE;
                for (k=0;k<D;k++) {
                   if ((oNode->Entries[j]).ver[k] < rg->start[k] 
                        || (oNode->Entries[j]).ver[k] > rg->end[k] ) {                       
                       result=FALSE;
                       break;
                   }
                }                                     
             /*   preOffset = oNode->Entries[j].Offset; */
                if (result) {                 
                    InsertList(resultList, oNode->Entries[j].Offset);
                    listLen++;
                    got++;
                }   
            }
            overPtr = oNode->Next;
        }
    }    
  }
  if (Continue &&  currNode->Entries[currNode->NumEntry-1].Key <= key2) {
    nextPtr = currNode->Next;
    if (nextPtr < 0) {
      Continue = FALSE;
    }
  } else {
    Continue = FALSE;

  }
} /* while */

  
  free(currNode);
  free(oNode);

  if (listLen > 0) {
    return (TRUE);
  } else {
    return (FALSE);
  }
}
