/* Version 1.0 (Beta Test)
 * Copyright (c) Rui Zhang, 2007
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The code will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - All bug fixes will be returned to rui@csse.unimelb.edu.au for
 *    inclusion in future releases
 */

/*
A program to divide the data into subspaces for P+-tree.
As a result, the data in subspaces are stored in files data0, 
data1, data2, ... etc. The space-tree is stored in the file "SpaceTree".
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>

#define D 16 //dimensionality

char* itoa(int i);
float distance(float *p, float *q);

typedef struct {
	int dim;
	float division;
	long leftchild;
	long rightchild;
}C_INODE;

typedef struct {
	int fileno;
	float a[D];
	float b[D];
	float c[D];
}C_LNODE;

main(int argc, char **argv)
{
	int T; //The order of division. There are 2^T subspaces
	
	if(argc==2)
	{
	T=atoi(argv[1]);
	FILE *fp_m, *fp_ml, *fp_mr, *fp_CTree;

	float *w,*p,*wl,*wr,*cl,*cr,border_o[4096][D][2], border_n[4096][D][2]; //w points to centroid, p points to a data point. wl and wr point to centroids of sub-clusters Ml and Mr; cl and cr point to the temporary centers for Ml and Mr, border stores min and max of each subcluster, 1024 means T<=10
	float a,b,median;
	long n,nl,nr, CToffset, CTvalue; //n is the number of points, nl and nr are number of points of Ml and Mr
	int cdt; //cdt is current dividing time
	int i,j,k,nameflag=0,clusterflag,clusterno;//clno is for inserting CTree leafnodes
	char m[20], ml[20], mr[20]; //m is the mother cluster file name, ml and mr are the subcluster file name
	C_INODE cinode;
	C_LNODE clnode;
	
	clock_t t1,t2;
	time_t time1,time2;
	
	t1=clock();
	time1=time(NULL);

	
	if((fp_CTree=fopen("SpaceTree","w+"))==NULL)
	{
		printf("can't create file CTree");
		exit(0);
	}


//the first long in CTreeFile points to the start of C_LNODE
	CTvalue=sizeof(long)+(pow(2,T)-1)*sizeof(C_INODE);
	fwrite(&CTvalue, sizeof(long), 1, fp_CTree);

	w=(float *)malloc(sizeof(float)*D);
	p=(float *)malloc(sizeof(float)*D);
	wl=(float *)malloc(sizeof(float)*D);
	wr=(float *)malloc(sizeof(float)*D);
	cl=(float *)malloc(sizeof(float)*D);
	cr=(float *)malloc(sizeof(float)*D);

	for(i=0;i<pow(2,T);i++)
	{
		for(j=0;j<D;j++)
		{
			border_o[i][j][0]=border_n[i][j][0]=0.0;
			border_o[i][j][1]=border_n[i][j][1]=1.0;
		}
	}

	for(cdt=0;cdt<T;cdt++)
	{
		for(k=0;k<pow(2,cdt);k++)
		{
			strcpy(m,"dat");
			strcpy(ml,"dat");
			strcpy(mr,"dat");
//produce file name for mother cluster
			if(nameflag==0)	strcat(m,"a");
			else strcat(m,"b");
			strcat(m,itoa(k));
//End produce file name for mother cluster
				
//produce file name for sub cluster ml
			if(nameflag==0)	strcat(ml,"b");
			else strcat(ml,"a");
			strcat(ml,itoa(k*2));
//End produce file name for sub cluster ml			
	
//produce file name for sub cluster ml
			if(nameflag==0)	strcat(mr,"b");
			else strcat(mr,"a");
			strcat(mr,itoa(k*2+1));
//End produce file name for sub cluster ml	
					
	
			if((fp_m=fopen(m,"r"))==NULL)
			{
				printf("can't open file %s\n",m);
				exit(0);
			}
	
			if((fp_ml=fopen(ml,"w+"))==NULL)
			{
				printf("can't create file ml");
				exit(0);
			}

			if((fp_mr=fopen(mr,"w+"))==NULL)
			{
				printf("can't open file mr");
				exit(0);
			}

//calculate centroid w
			for(i=0;i<D;i++) w[i]=0;

//				while(fread(p,sizeof(float),D,fp_m))
//				{
//					n++;
//					for(i=0;i<D;i++) w[i]+=p[i];
//				}
			n=0;
			
			while(fread(p,D*sizeof(float),1,fp_m))
			{
				n++;
				for(i=0;i<D;i++) w[i]+=p[i];
			}

			for(i=0;i<D;i++) w[i]/=n;
			for(i=0;i<D;i++) printf("%f ",w[i]);
			printf("\n");
//choose the dimension where w's ordinate is nearest to median
//then determin cl and cr according to this dimension
/*				a=fabs(w[0]-(border_o[k][0][0]+border_o[k][0][1])*0.5);
				j=0;
				for(i=1;i<D;i++)
				{
					if(fabs(w[i]-(border_o[k][i][0]+border_o[k][i][1])*0.5)<a)
					{
						a=fabs(w[i]-(border_o[k][i][0]+border_o[k][i][1])*0.5);
						j=i;
					}
				}
		
				for(i=0;i<D;i++) cl[i]=cr[i]=w[i];
				if(w[j]<(border_o[k][j][0]+border_o[k][j][1])*0.5)
				{
					cl[j]=(w[j]+border_o[k][j][0])*0.5;
					cr[j]=2*w[j]-cl[j];
				}
				else
				{
					cr[j]=(w[j]+border_o[k][j][1])*0.5;
					cl[j]=2*w[j]-cr[j];
				}
*/
				
//another way to determin cl and cr
			for(i=0;i<D;i++)
			{
				if(w[i]<(border_o[k][i][0]+border_o[k][i][1])*0.5)
				{
					cl[i]=(w[i]+border_o[k][i][0])*0.5;
					cr[i]=2*w[i]-cl[i];
				}
				else
				{
					cr[i]=(w[i]+border_o[k][i][1])*0.5;
					cl[i]=2*w[i]-cr[i];
				}
			}

//calculate wl, wr according to cl, cr, w until wl=cl and wr=cr
			do
			{
				rewind(fp_m);

				nl=nr=0;
				for(i=0;i<D;i++) wl[i]=wr[i]=0;
	
				while(fread(p,D*sizeof(float),1,fp_m))
				{
					if( distance(p,cl)<distance(p,cr) )
					{
						nl++;
						for(i=0;i<D;i++) wl[i]+=p[i];
					}
					else
					{
						nr++;
						for(i=0;i<D;i++) wr[i]+=p[i];
					}
				}
				
				for(i=0;i<D;i++)
				{
					wl[i]/=nl;
					wr[i]/=nr;
				}
		//check if wl==cl and wr==cr
				for(i=0;i<D && wl[i]==cl[i];i++);
				for(j=0;j<D && wr[j]==cr[j];j++);

				if(i==D && j==D) clusterflag=0;
				else 
				{
					clusterflag=1;
					for(i=0;i<D;i++)
					{
						cl[i]=wl[i];
						cr[i]=wr[i];
					}
				}

			}while(clusterflag);

//find the dimension where cl differs from cr most
			a=fabs(cl[0]-cr[0]);
			j=0;
			for(i=1;i<D;i++)
			{
				if(fabs(cl[i]-cr[i])>a)
				{
					a=fabs(cl[i]-cr[i]);
					j=i;
				}
			}
			median=(cl[j]+cr[j])/2.0;
				
//write C_INODE
			CToffset=ftell(fp_CTree);

			cinode.dim=j;
			cinode.division=median;

			if(cdt<T-1)
			{
				cinode.leftchild=CToffset+(pow(2,cdt)+k)*sizeof(C_INODE);
				cinode.rightchild=CToffset+(pow(2,cdt)+k+1)*sizeof(C_INODE);
			}
			else
			{
				cinode.leftchild=CToffset+(pow(2,cdt)-k)*sizeof(C_INODE)+2*k*sizeof(C_LNODE);
				cinode.rightchild=CToffset+(pow(2,cdt)-k)*sizeof(C_INODE)+(2*k+1)*sizeof(C_LNODE);
			}

			
			fwrite(&cinode,sizeof(C_INODE),1,fp_CTree);

//End write C_INODE

//write to mlFile, mrFile
			rewind(fp_m);

			while(fread(p,D*sizeof(float),1,fp_m))
			{
				if( p[j]<median )
				{
					fwrite(p,D*sizeof(float),1,fp_ml);
				}
				else
				{
					fwrite(p,D*sizeof(float),1,fp_mr);
				}
			}

			for(i=0;i<D;i++) 
			{
				border_n[k*2][i][0]=border_o[k][i][0];
				border_n[k*2][i][1]=border_o[k][i][1];

				border_n[k*2+1][i][0]=border_o[k][i][0];
				border_n[k*2+1][i][1]=border_o[k][i][1];
			}
				
			border_n[k*2][j][1]=median;
			border_n[k*2+1][j][0]=median;
//End write to mlFile, mrFile



				
			fclose(fp_m);
			fclose(fp_ml);
			fclose(fp_mr);


			remove(m);

		}

		for(i=0;i<pow(2,T);i++)
		{
			for(j=0;j<D;j++)
			{
				border_o[i][j][0]=border_n[i][j][0];
				border_o[i][j][1]=border_n[i][j][1];
			}
		}				
		nameflag=1-nameflag;
	}

		//printf("\n%d  %d\n",CTvalue,CTreeFile.GetPosition());

//write C_LNODE


	for(clusterno=0;clusterno<pow(2,T);clusterno++)
	{
		strcpy(m,"dat");
		if(nameflag==0)	strcat(m,"a");
		else strcat(m,"b");
		strcat(m,itoa(clusterno));
		
		
		if((fp_m=fopen(m,"r"))==NULL)
		{
			printf("can't open file %s when building CTree\n",m);
			exit(0);
		}		
		
			
		for(i=0;i<D;i++) w[i]=0;
		n=0;
		
		while(fread(p,D*sizeof(float),1,fp_m))
		{
			n++;
			for(i=0;i<D;i++) w[i]+=p[i];
		}
		
		for(i=0;i<D;i++) w[i]/=n;
		//calculate c[i]
		printf("clusterNo. dimension lowerborder higherborder clustercenter\n");
		for(i=0;i<D;i++)
		{
			a=1/(border_n[clusterno][i][1]-border_n[clusterno][i][0]);
			b=border_n[clusterno][i][0]/(border_n[clusterno][i][1]-border_n[clusterno][i][0]);
			clnode.a[i]=a;
			clnode.b[i]=b;
			clnode.c[i]=-log(2)/log(a*w[i]-b);

			printf("%d %d %f %f %f\n",clusterno,i,border_n[clusterno][i][0],border_n[clusterno][i][1],w[i]);

		}


		clnode.fileno=clusterno;
			
		//write clnode
		fwrite(&clnode,sizeof(C_LNODE),1,fp_CTree);	
		
		fclose(fp_m);
	}
//End write C_LNODE
	
	fclose(fp_CTree);
	t2=clock();
	time2=time(NULL);
	
	printf("\n cluster end\nClock:%ld",t2-t1);
	printf("Time: %f\n",difftime(time2,time1));

	//getchar();
	}
	else printf("Input: OrderOfDivision\n\nCurrent D=%d\n",D);
}
		
char* itoa(int i)
{
	int j,k=0,l;
	char *a, tp;
	a=(char *)malloc(sizeof(char)*20);
	if(i>=0)
	{
		do
		{
			j=i%10;
			a[k]=j+48;
			k++;
			i=i/10;
		}while(i>0);
	}
//now a is the inverse of i, so reinverse it.	
	for(l=0;l<k/2;l++)
	{
		tp=a[l];
		a[l]=a[k-1-l];
		a[k-1-l]=tp;
	}

	a[k]=0;
	return a;
}

float distance(float *p, float *q)
{
	float dis=0;
	int i;
	for (i=0;i<D;i++) dis+=(p[i]-q[i])*(p[i]-q[i]);
	return pow(dis,0.5);
}
