/* Version 1.0 (Beta Test)
 * Copyright (c) Rui Zhang, 2007
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The code will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - All bug fixes will be returned to rui@csse.unimelb.edu.au for
 *    inclusion in future releases
 */

/*
The P+-tree main program.
Use the command "pplus b" to build index and "pplus" to
perform window queries.
Dimensionality "D" is defined in btree.h; order of division
of the space is defined by "T" below.
*/

#include "stdio.h"
#include "stdlib.h"
#include "math.h"
#include "btree.h"
#include "nulllist.h"

#include "time.h"

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define T 4 //The order of division. There are 2^T subspaces

typedef struct {
    int pyramid ;
    float height ;
} pyrval ;

typedef struct {
	int dim;
	float division;
	long leftchild;
	long rightchild;
}C_INODE;

typedef struct {
	int fileno;
	float a[D];
	float b[D];
	float c[D];
}C_LNODE;



extern int BT_Insert(BTREE *, const long, const KEYTYPE, RECORD_TYPE *);
extern int BT_Open(BTREE *, FILE *, PTR, PTR);
extern int BT_Close(BTREE *);

extern PTR BT_Locate(BTREE *, PTR, const KEYTYPE);
extern int BT_NewLNode(LNODE **);
extern int BT_IOReadLNode(BTREE *, PTR, LNODE *);
extern void IO_ResetPageAccess(void);

extern int BT_NewONode(ONODE **);
extern int BT_IOReadONode(BTREE *, PTR, ONODE *);

extern IMINODE* LoadINodes(PTR , BTREE *);
extern PTR IMBT_Locate(IMINODE *, KEYTYPE);
extern IMINODE* IMBT_LocateBaseINode(IMINODE *piminode, KEYTYPE key);

extern int QueueSize;
extern long PageAccess;
int got=0;
int candidate=0;
extern int max_top=-1;

FILE *fp_result, *fp_index;
BTREE btree;
float oriqw[D][2]={{0.0,0.4},{0.2,0.6},{0.3,0.7},{0.6,1.0}};//query window

IMINODE *root;

long NoAnswer=0,TotalNoAnswer=0;

char* itoa(int i);
float pv(float * vector);
void CTSearch(long ctoffset, long ctleafstart, FILE *fp_ct);


float MIN(float *r);
float MAX(float *r);
int intersect(float qw[D][2], int i, float *hlow, float *hhigh);
void DeterminRange(float qw[D][2], int i, float *hlow, float *hhigh);
void RangeSearchLeaf(BTREE *pbtree, PTR pid, float lower, float higher, float qw[D][2]);
void RangeSearchOFLeaf(BTREE *pbtree, PTR pid, float lower, float higher, float qw[D][2]);
void CheckRange(LENTRY *le, float qw[D][2]);




main(int argc, char ** argv)
{
	FILE *fp_m, *fp_CTree, *fp_query, *fp_btreeinfo;
	char m[20];
	int fileno,i;
	float *p,base;
	KEYTYPE pvalue;
	long offset=0L, CToffset, CTleafstart, CTcur;
	RECORD_TYPE record;
	C_LNODE clnode;

	long TotalPage,NumQuery;
	
	clock_t t1,t2;
	time_t time1,time2;

//build btree	
if(argc==2 && strcmp (argv [1], "b") == 0)
{	
	//p=(float *)malloc(sizeof(float)*D);
	
	if((fp_index=fopen("index","w+"))==NULL)
	{
		printf("can't create index file");
		exit(0);
	}

	if((fp_CTree=fopen("cluster/SpaceTree","r"))==NULL)
	{
		printf("can't open file SpaceTree");
		exit(0);
	}
		
	BT_Open(&btree,fp_index,-1,-1);
	
	t1=clock();
	time1=time(NULL);
	
	fread(&CTleafstart, sizeof(long),1,fp_CTree);
	
	p=(float *)malloc(sizeof(float)*D);
		
	
	for(fileno=0;fileno<pow(2,T);fileno++)
	{
		if(T%2==0)	strcpy(m,"cluster/data");
		else strcpy(m,"cluster/datb");
		strcat(m,itoa(fileno));
		
		fseek(fp_CTree,CTleafstart+fileno*sizeof(C_LNODE),SEEK_SET);
		fread(&clnode,sizeof(C_LNODE),1,fp_CTree);
		
		if(fileno!=clnode.fileno)
		{
			printf("wrong fileno\n");
			exit(0);
		}
		
		if((fp_m=fopen(m,"r"))==NULL)
		{
			printf("can't open file m %d",fileno);
			exit(0);
		}
		
		base=2*D*fileno;
		
		//p=(float *)malloc(sizeof(float)*D); //delete if right
		
		while(fread(&record,D*sizeof(float),1,fp_m))
		{
			for(i=0;i<D;i++)
			{
				p[i]=pow(record.fields[i]*clnode.a[i]-clnode.b[i],clnode.c[i]);
			}
			pvalue=(KEYTYPE)pv(p)+(KEYTYPE)base;
			//for(i=0;i<D;i++) record.
			BT_Insert(&btree,offset,pvalue,&record);
			offset++;
		}

		fclose(fp_m);
	}
	
	t2=clock();
	time2=time(NULL);

	//save btreeinfo information
	fp_btreeinfo=fopen("btreeinfo","w+");

	fprintf (fp_btreeinfo, "%ld\n", btree.Root);
	fprintf (fp_btreeinfo, "%ld\n", btree.Leaf_Head);

	fclose(fp_btreeinfo);

	
	printf("Root: %d\n",btree.Root);
 	printf("Left most leaf node: %d \n", btree.Leaf_Head);
 	printf("Building index clock: %ld\n Time: %f\n",t2-t1,difftime(time2,time1));
 	BT_Close(&btree);
 	fclose(fp_CTree);	
}
//End build btree
else
{
	if((fp_CTree=fopen("cluster/SpaceTree","r"))==NULL)
	{
		printf("can't open file SpaceTree");
		exit(0);
	}
	
	if((fp_result=fopen("result","w+"))==NULL)
	{
		printf("can't create file result");
		exit(0);
	}
	
	if((fp_index=fopen("index","r"))==NULL)
	{
		printf("can't open file index");
		exit(0);
	}
	
	if((fp_query=fopen("query/window","r"))==NULL)
	{
		printf("can't open file window");
		exit(0);
	}

	//get the btree information
	if((fp_btreeinfo=fopen("btreeinfo","r"))==NULL)
	{
		printf("can't open file btreeinfo");
		exit(0);
	}
	fscanf (fp_btreeinfo, "%ld", &(btree.Root));
      	fscanf (fp_btreeinfo, "%ld", &(btree.Leaf_Head));		
	fclose(fp_btreeinfo);
	
	
	TotalPage=NumQuery=0;
	
	BT_Open(&btree,fp_index,btree.Root,btree.Leaf_Head);
	
	//load internal nodes of BTree
	
	//view root info
	//INODE *root;
	//BT_NewINode(&root);
	//BT_IOReadINode(&btree, btree.Root, root);
	//printf("information of root:\n");
	//printf("Level:%d, Number of Key in Root: %d \n I_FAN_OUT: %d, L_FAN_OUT: %d \n",root->Level,root->NumKey,I_FAN_OUT,L_FAN_OUT);
	//End view root info
	
	root=LoadINodes(btree.Root, &btree);	
	printf("INODES Load Over\n");
	//End load internal nodes of BTree
	
	t1=clock();
	time1=time(NULL);
	
	while(fread(oriqw,sizeof(float),2*D,fp_query))
	{
		IO_ResetPageAccess();
		NoAnswer=0;
		
		rewind(fp_CTree);
	
		fread(&CTleafstart, sizeof(long),1,fp_CTree);
			
		CTcur=ftell(fp_CTree);
	
		CTSearch(CTcur,CTleafstart,fp_CTree);
		
		for(i=0;i<D;i++)
		{
			fprintf(fp_result,"%f %f,",oriqw[i][0],oriqw[i][1]);
		}
		
		fprintf(fp_result,"\nPage Access: %ld   ",PageAccess);
		fprintf(fp_result,"NoAnswer: %ld\n",NoAnswer);
		
		TotalPage+=PageAccess;
		NumQuery++;
		TotalNoAnswer+=NoAnswer;
	}
	
	t2=clock();
	time2=time(NULL);
	
	TotalPage=TotalPage/NumQuery;
	TotalNoAnswer=TotalNoAnswer/NumQuery;
	
	fprintf(fp_result,"\nAverage Page Access:%ld",TotalPage);
	fprintf(fp_result,"\nAverage NoAnswer:%ld",TotalNoAnswer);
	fprintf(fp_result,"\nClock:%ld  ",t2-t1);
	fprintf(fp_result,"Time: %f\n",difftime(time2,time1));
		
	fclose(fp_query);
	
	BT_Close(&btree);
	fclose(fp_result);
	fclose(fp_CTree);	
	
}

}

char* itoa(int i)
{
	int j,k=0,l;
	char *a, tp;
	a=(char *)malloc(sizeof(char)*20);
	if(i>=0)
	{
		do
		{
			j=i%10;
			a[k]=j+48;
			k++;
			i=i/10;
		}while(i>0);
	}
//now a is the inverse of i, so reinverse it.	
	for(l=0;l<k/2;l++)
	{
		tp=a[l];
		a[l]=a[k-1-l];
		a[k-1-l]=tp;
	}

	a[k]=0;
	return a;
}

float pv(float * vector) //pyramid value
{
    pyrval p = {-1, -1.0} ;
    int j ;
    for (j=0 ; j<D ; j++)
		if (vector [j] <= 0.5) {
			if (p.height < 0.5 - vector [j]) {
				p.height = 0.5 - vector [j] ;
				p.pyramid = j ;
			}
		} else {
			if (p.height < vector [j] - 0.5) {
				p.height = vector [j] - 0.5 ;
				p.pyramid = j + D;
			}
		}
		return p.height+p.pyramid;
}

void CTSearch(long ctoffset, long ctleafstart, FILE *fp_ct)
{
	C_INODE cinode;
	C_LNODE clnode;
	float subqw[D][2];//subquery window
	int i,j,fileno;
	float hlow,hhigh;
	float qw[D][2];
	PTR pid;
	
	fseek(fp_ct,ctoffset,SEEK_SET);
	//If this node is C_INODE
	if(ctoffset<ctleafstart)
	{
		fread(&cinode,sizeof(C_INODE),1,fp_ct);
		
		//intersect left child
		if(oriqw[cinode.dim][0]<=cinode.division) CTSearch(cinode.leftchild,ctleafstart,fp_ct);
		
		//intersect right child
		if(oriqw[cinode.dim][1]>=cinode.division) CTSearch(cinode.rightchild,ctleafstart,fp_ct);
	}
	
	//If this node is C_LNODE
	else
	{
		fread(&clnode,sizeof(C_LNODE),1,fp_ct);
		
		//construct subquery
		for(i=0;i<D;i++)
		{
			if(clnode.a[i]*oriqw[i][0]-clnode.b[i]<=0) subqw[i][0]=0;
			else	subqw[i][0]=pow(clnode.a[i]*oriqw[i][0]-clnode.b[i],clnode.c[i]);
			//if (subqw[i][0]<0) subqw[i][0]=0;
			subqw[i][1]=pow(clnode.a[i]*oriqw[i][1]-clnode.b[i],clnode.c[i]);
			if (subqw[i][1]>1) subqw[i][1]=1;
		}
		
		fileno=clnode.fileno;
		
		//pyramid query
		for(i=0;i<D;i++)
		{
			subqw[i][0]-=0.5;
			subqw[i][1]-=0.5;
		}
		
		for(i=0;i<2*D;i++)
		{
			for(j=0;j<D;j++) 
			{
				qw[j][0]=subqw[j][0];
				qw[j][1]=subqw[j][1];
			}
		
			if(intersect(qw,i,&hlow,&hhigh))
			{
				//BTree range query
				//pid=BT_Locate(&btree,btree.Root,2*fileno*D+i+hlow); //for disk BTree
				pid=IMBT_Locate(root,2*fileno*D+i+hlow);
				RangeSearchLeaf(&btree,pid,2*fileno*D+i+hlow,2*fileno*D+i+hhigh,qw);
			}
		}		
	}
}

float MIN(float *r)
{
	if(r[0]*r[1]<=0) return 0;
	else return fabs(r[0])<fabs(r[1])? fabs(r[0]):fabs(r[1]);
}

float MAX(float *r)
{
	return fabs(r[0])>fabs(r[1])? fabs(r[0]):fabs(r[1]);
}

int intersect(float qw[D][2], int i, float *phlow, float *phhigh)
{
	int j;
	if(i<D)
	{
		for(j=0;j<D && qw[i][0]<=-MIN(qw[j]);j++);
		if(j==D)
		{
			if (qw[i][1]>0) qw[i][1]=0;
			DeterminRange(qw,i,phlow,phhigh);
			return 1;
		}
		else return 0;
	}
	
	if(D-1<i)
	{
		for(j=0;j<D && qw[i-D][1]>=MIN(qw[j]);j++);
		if(j==D)
		{
			if (qw[i-D][0]<0) qw[i-D][0]=0;
			DeterminRange(qw,i-D,phlow,phhigh);
			return 1;
		}
		else return 0;
	}
}

void DeterminRange(float qw[D][2], int i, float *phlow, float *phhigh)
{
	int j,k;
	float qtemp;
	*phhigh=MAX(qw[i]);
	for(j=0;j<D && qw[j][0]<=0 && 0<=qw[j][1]; j++);
	if(j==D) *phlow=0;
	else
	{
		*phlow=0.0;
		for(k=0;k<D;k++)
		{
			qtemp=MIN(qw[k])>MIN(qw[i])? MIN(qw[k]):MIN(qw[i]);
			if(qtemp > *phlow) *phlow=qtemp;
		}
	}
}

void RangeSearchLeaf(BTREE *pbtree, PTR pid, float lower, float higher, float qw[D][2])
{
	LNODE *lnode;
	int i;
	
	BT_NewLNode(&lnode);
	BT_IOReadLNode(pbtree,pid,lnode);
	
	if(lnode->NumEntry!=0)
	{
		for(i=0;i<lnode->NumEntry;i++)
		{
			if(lower<=lnode->Entries[i].Key && lnode->Entries[i].Key<=higher)
			{
				CheckRange(&(lnode->Entries[i]),qw);
			}
		}
		
		if(lnode->Overflow>0) RangeSearchOFLeaf(pbtree,lnode->Overflow,lower,higher,qw);
		
		if(lnode->Entries[lnode->NumEntry-1].Key<=higher && lnode->Next>=0)
			RangeSearchLeaf(pbtree,lnode->Next,lower,higher,qw);
	}

	else if(lnode->Next>=0)
	{
		RangeSearchLeaf(pbtree,lnode->Next,lower,higher,qw);
	}
	
	free(lnode);
}

void RangeSearchOFLeaf(BTREE *pbtree, PTR pid, float lower, float higher, float qw[D][2])
{
	ONODE *onode;
	int i;
	
	BT_NewONode(&onode);
	BT_IOReadONode(pbtree,pid,onode);
	
	if(onode->NumEntry!=0)
	{
		for(i=0;i<onode->NumEntry;i++)
		{
			if(lower<=onode->Entries[i].Key && onode->Entries[i].Key<=higher)
			{
				CheckRange(&(onode->Entries[i]),qw);
			}
		}
		
		if(onode->Next>0) RangeSearchOFLeaf(pbtree,onode->Next,lower,higher,qw);
		
		//if(onode->Entries[onode->NumEntry-1].Key<=higher && onode->Next>=0)
			//RangeSearchOFLeaf(pbtree,onode->Next,lower,higher,qw);
	}

	else if(onode->Next>=0)
	{
		RangeSearchOFLeaf(pbtree,onode->Next,lower,higher,qw);
	}
	
	free(onode);
}

void CheckRange(LENTRY *le, float qw[D][2])
{
	int i,j;
	for(i=0; i<D && oriqw[i][0]<=le->ver[i] && le->ver[i]<=oriqw[i][1]; i++);
	if (i==D) 
	{
		NoAnswer++;
		
		//fprintf(fp_result,"%ld ",le->Offset);
		
		for(j=0;j<D;j++) fprintf(fp_result,"%f ",le->ver[j]);
		fprintf(fp_result,"\n");
	}
}

